# ../addons/eventscripts/mpbhops_es/mpbhops_es.py
# mpbhops in ES/python
# written by Absolute
# Special Thanks to:
# - L'In20Cible
# - DaFox


import es
import os
from time import time
import effectlib
import vecmath
import cfglib
import spe
from spe import HookType
from spe import HookAction
from spe import Conventions


rehook = 0
# ignore = []
teleporterlist = []
bhopdict = {}
playerdict = {}
maxplayers = es.getmaxplayercount()


# Offset for CBaseEntity::Touch(CBaseEntity*)
OFFSET_TOUCH = 98 if spe.platform == "nt" else 99
OFFSET_VECPOSITION1 = 840 if spe.platform == "nt" else 860
OFFSET_VECPOSITION2 = 852 if spe.platform == "nt" else 872
#OFFSET_LOCKED = 1636 if spe.platform == "nt" else 1656
FUNC_START_TOUCH = None
MAXDISTANCE = 48.0 # maximum distance a platform can be above a trigger


Button_touch = 1<<8 # Opens on Touch
Door_touch = 1<<10 # Opens on Touch

info = es.AddonInfo()
info.author      = "Absolute"
info.name        = "Mpbhops"
info.basename    = "mpbhops_es"
info.version     = "2.0"
info.url         = "http://addons.eventscripts.com/addons/view/%s"%info.basename
info.description = "Stops the platforms from going down" 

es.ServerVar(info.basename, info.version, info.description + " made by " + info.author).makepublic()

# =============================
# Configlib
# =============================

config = cfglib.AddonCFG( os.path.join( es.getAddonPath(info.basename), "config.cfg" ) )
config.text("../addons/eventscripts/%s/config.cfg" % info.basename )
config.text("%s Version %s by %s" % ( info.name, info.version, info.author ) )
config.text("For more info, visit: %s" % info.url)
colorHooked = config.cvar("colorHooked", 0, "If this is enabled (1) then it will color all hooked platforms/doors")
waittime = config.cvar("waittime", 0.1, "After how many seconds it will start to teleport you if you touch the same platform again")
resettime = config.cvar("resettime", 1, "After how many seconds you can touch any platform again")

config.write()
config.execute()
waittime = float(waittime)
resettime = float(resettime)


# EVENTS

def load():
    # for late load, hook platforms and trigger the events
    for x in es.getUseridList():
        player_activate({"userid":x})
    setupPlatforms()
    round_start(0) 

def unload():
    # Reset dicts
    bhopdict.clear()
    playerdict.clear()
    unhookStartTouch(pre_touch)


def player_spawn(event_var):
    if rehook:
        disabledoors(event_var["userid"])


def round_start(event_var):
    setupPlatforms()
    x = es.getuserid()
    if x:
        disabledoors(x)
    else:
        global rehook
        rehook = 1
    

def player_activate(ev):
    userid = int(ev["userid"])
    playerindex = es.getindexfromhandle(es.getplayerhandle(userid))
    playerdict[playerindex] = {}
    playerdict[playerindex]['lastindex'] = 0
    playerdict[playerindex]['lasttouched'] = 0
    
def player_disconnect(ev):
    userid = int(ev["userid"])
    playerindex = es.getindexfromhandle(es.getplayerhandle(userid))
    if playerindex in playerdict:
        del playerdict[playerindex]
        
    
def pre_touch(args):
    '''
    Called before an entity touches another entity.
    '''
    if len(args) != 2: return (HookAction.Continue, 0)
    playerindex = spe.getIndexOfEntity(args[0])
    bhop = spe.getIndexOfEntity(args[1])
#    es.msg("player %s, %i" % (playerindex, bhop))
    if playerindex < maxplayers and playerindex >= 0 and bhop in bhopdict: 
#    userid = es.getuserid(playerindex)
#    es.msg("player %s, %i" % (playerindex, bhop))
#    
        if bhop != playerdict[playerindex]['lastindex']:
            playerdict[playerindex]['lastindex'] = bhop
            playerdict[playerindex]['lasttouched'] = time()
        else:
            lastTouch = time()-playerdict[playerindex]['lasttouched']
            if lastTouch > resettime:
                playerdict[playerindex]['lastindex'] = 0
                playerdict[playerindex]['lasttouched'] = 0
            elif lastTouch > waittime:
                playerdict[playerindex]['lastindex'] = 0
                playerdict[playerindex]['lasttouched'] = 0
                bhopdict[bhop].teleportToTarget(es.getuserid(es.gethandlefromindex(playerindex)))
            
    return (HookAction.Continue, 0)
    

def unhookStartTouch(callback):
    if FUNC_START_TOUCH != None:
        spe.unHookFunction(FUNC_START_TOUCH, HookType.Pre, callback)

    
    
def setupPlatforms():
    """ 
    Hooks the platforms
    """
        
    # Reset dicts
    bhopdict.clear()
    global teleporterlist
    teleporterlist = []
    
    global FUNC_START_TOUCH
    
    # Get teleports
    for key in es.createentityindexlist("trigger_teleport"):
        teleporterlist.append(Teleporter(key))
        
                
    for key in (es.createentityindexlist("func_button").keys() + es.createentityindexlist("func_door").keys()):
        platform = Platform(key)
        if platform.classname == "func_button" and not platform.spawnflags & Button_touch: continue
        if platform.classname == "func_door"   and not platform.spawnflags & Door_touch: continue
        ent = spe.getEntityOfIndex(key)
        
        if FUNC_START_TOUCH == None:
            # hook the start touch function
            FUNC_START_TOUCH = spe.findVirtualFunc(ent, OFFSET_TOUCH)
            spe.hookFunction(FUNC_START_TOUCH, 'pp)v', Conventions.THISCALL, HookType.Pre, pre_touch)
        
        vec1 = spe.getLocVal('f', ent+OFFSET_VECPOSITION1+8) # get the z coordinate of vecposition1
        vec2 = spe.getLocVal('f', ent+OFFSET_VECPOSITION2+8) # get the z coordinate of vecposition2
        if vec1 > vec2: # the platform moves downwards
            if platform.teleporter:
                bhopdict[key] = platform
                platform.setKeyValue("targetname", "hooked")
    global rehook
    rehook = 1
                    
def disabledoors(uid):
    global rehook
    rehook = 0
    es.server.queuecmd('es_fire %s hooked addoutput "speed 0"' % uid)
    es.server.queuecmd('es_fire %s hooked Lock'% uid)
    if colorHooked:
        es.server.queuecmd('es_fire %s hooked Color "255 0 0"' % uid)
        es.server.queuecmd('es_fire %s buthooked Color "255 0 0"' % uid)
    es.server.queuecmd('es_fire %s hooked addoutput "noise1 doors/default_move.wav"'%uid)
    


# CLASSES

class Entity(object):
    def __init__(self, index):
        self.index = index
        self.origin = self.getVectorProp("CBaseEntity.m_vecOrigin")
        self.vecMaxs = self.getVectorProp("CBaseEntity.m_Collision.m_vecMaxs")
        self.vecMins = self.getVectorProp("CBaseEntity.m_Collision.m_vecMins")
        self.classname = self.getKeyValue("classname")
        
    @property
    def spawnflags(self):
        return int(self.getKeyValue("spawnflags"))
    
    def getKeyValue(self, name):
        return es.entitygetvalue(self.index, name)
        
    def setKeyValue(self, name, value):
        es.entitysetvalue(self.index, name, value)
        
    
        
    def getProp(self, prop):
        return es.getindexprop(self.index, prop)
        
    def getVectorProp(self,prop):
        return vecmath.vector(self.getProp(prop))

class Teleporter(Entity):
    def __init__(self, index):
        super(Teleporter, self).__init__(index)
        self.minvector = self.origin+self.vecMins
        self.maxvector = self.origin+self.vecMaxs
        
    def teleport(self, userid):
        es.server.queuecmd("es_setpos %s %s" % (userid, self.origin.getstr(" ")))
        
    def isValidTeleporter(self, platform):
        startpoint = platform.origin
        return (startpoint[2]-self.maxvector[2] <= MAXDISTANCE and
                all(startpoint[x] <= self.maxvector[x] for x in range(2)) and 
                all(startpoint[x] >= self.minvector[x] for x in range(3))) # check if the platform is above the trigger, but still inside it
    
    def drawTeleporter(self):
        effectlib.drawBox(self.maxvector, self.minvector, seconds=15, red=255, green=0, blue=0)

        
class Platform(Entity):
    def __init__(self, index):
        super(Platform, self).__init__(index)
        self.__teleporter = None
        
    def teleportToTarget(self, userid):
        self.teleporter.teleport(userid)
        
    @property
    def teleporter(self):
        if self.__teleporter != None:
            return self.__teleporter
        for tele in teleporterlist:
            if tele.isValidTeleporter(self):
                self.__teleporter = tele
                return self.teleporter
        return None
        
